.globl _start
.set noat
.set noreorder

.text

;
;;;
;;; Sorts 16 integers at address 0x80 using Bubble Sort
;;;
;;; This is for the (still) unsafe optimized pipelined MIPS data path
;;; with some forwarding implemented.
;;;
;;; The "beq" instruction is processed in the ID stage, therefore the "nop"
;;; instructions after "beq" only need to fill one branch delay slot.
;;;
;;; The register file supports bypassing, i.e., whatever is being written
;;; can be read at the same. The ID, EX, and MA stages also implement some
;;; form of forwarding, which allows removing most of the "nop" instruction.
;;; The ones left avoid the hazards due to load-use dependency, as also the
;;; hazard due to dependency between "beq" and previous instructions, because
;;; "beq" being executed in the ID stage means that it is executed in parallel
;;; with the instruction in EX stage which can provide a result for "beq".
;;;
;

_start:
	addi  $a0, $zero, 16		; n = 16
	addi  $a1, $zero, 1		; one = 1 (we don't have slti yet)

outer_loop:
	slt   $t0, $a0, $a1		; $a0 < one?
	nop			; avoid data hazard
	beq   $t0, $a1, outer_end	; $t0 == 1 => yes, goto outer_end
	nop				; branch delay slot

	addi  $s0, $zero, 0x80		; ptr = 0x80
	add   $s1, $zero, $zero		; i = 0
	add   $s2, $zero, $zero		; new_n = 0

inner_loop:
	slt   $t0, $s1, $a0		; i < n?
	nop			; avoid data hazard
	beq   $t0, $zero, inner_end	; $t0 == 0 => no, goto inner_end
	nop				; branch delay slot

	lw    $v0, 0 ($s0)		; v0 = *ptr
	lw    $v1, 4 ($s0)		; v1 = *(ptr + 4)
	nop				; avoid data hazard
	slt   $t0, $v1, $v0		; v1 < v0?
	nop			; avoid data hazard
	beq   $t0, $zero, inner_next	; $t0 == 0 => v1 >= v0 branch to inner_next
	nop				; branch delay slot

	sw    $v1, 0 ($s0)		; swap
	sw    $v0, 4 ($s0)
	add   $s2, $zero, $s1		; new_n = i

inner_next:
	addi  $s0, $s0, 4		; ptr = ptr + 4
	addi  $s1, $s1, 1		; i = i + 1
	beq   $0, $0, inner_loop	; goto inner_loop (unconditional relative jump)
	nop				; branch delay slot

inner_end:
	add   $a0, $zero, $s2		; n = new_n
	beq   $0, $0, outer_loop	; goto outer_loop (unconditional relative jump)
	nop				; branch delay slot

outer_end:
	beq    $0, $0, outer_end	; goto outer_end (unconditional relative jump)
	nop				; branch delay slot
