.globl _start
.set noat
.set noreorder

.text

;
;;;
;;; Sorts 16 integers at address 0x80 using Bubble Sort
;;;
;;; This is for the unsafe (but optimized) pipelined MIPS data path.
;;;
;;; The "beq" instruction is processed one stage earlier, therefore
;;; the "nop" instructions after "beq" only need to fill one branch delay slot
;;;
;;; The register file supports bypassing, i.e., whatever is being written
;;; can be read at the same. Therefore writes to registers being read are
;;; visible one cycle earlier, therefore we need one less "nop" instruction
;;; to avoid data hazards.
;;;
;

_start:
	addi  $a0, $zero, 16		; n = 16
	addi  $a1, $zero, 1		; one = 1 (we don't have slti yet)
	nop				; avoid data hazard
	nop				; avoid data hazard

outer_loop:
	slt   $t0, $a0, $a1		; $a0 < one?
	nop				; avoid data hazard
	nop				; avoid data hazard
	beq   $t0, $a1, outer_end	; $t0 == 1 => yes, goto outer_end
	nop				; branch delay slot

	addi  $s0, $zero, 0x80		; ptr = 0x80
	add   $s1, $zero, $zero		; i = 0
	add   $s2, $zero, $zero		; new_n = 0
	nop				; avoid data hazard

inner_loop:
	slt   $t0, $s1, $a0		; i < n?
	nop				; avoid data hazard
	nop				; avoid data hazard
	beq   $t0, $zero, inner_end	; $t0 == 0 => no, goto inner_end
	nop				; branch delay slot

	lw    $v0, 0 ($s0)		; v0 = *ptr
	lw    $v1, 4 ($s0)		; v1 = *(ptr + 4)
	nop				; avoid data hazard
	nop				; avoid data hazard
	slt   $t0, $v1, $v0		; v1 < v0?
	nop				; avoid data hazard
	nop				; avoid data hazard
	beq   $t0, $zero, inner_next	; $t0 == 0 => v1 >= v0 branch to inner_next
	nop				; branch delay slot

	sw    $v1, 0 ($s0)		; swap
	sw    $v0, 4 ($s0)
	add   $s2, $zero, $s1		; new_n = i

inner_next:
	addi  $s0, $s0, 4		; ptr = ptr + 4
	addi  $s1, $s1, 1		; i = i + 1
	beq   $0, $0, inner_loop	; goto inner_loop (unconditional relative jump)
	nop				; branch delay slot

inner_end:
	add   $a0, $zero, $s2		; n = new_n
	beq   $0, $0, outer_loop	; goto outer_loop (unconditional relative jump)
	nop				; branch delay slot

outer_end:
	beq    $0, $0, outer_end	; goto outer_end (unconditional relative jump)
	nop				; branch delay slot
